# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <mkoehl@cs.uni-saarland.de>

from __future__ import annotations

import os
import pathlib
import shutil
import subprocess
import sys
import tarfile

from distutils.command.build_ext import build_ext  # type: ignore

from urllib import request

import cffi


DEPS_PATH = pathlib.Path("deps")


ONIGURUMA_VERSION = "6.9.4"

ONIGURUMA_URL = f"https://github.com/kkos/oniguruma/archive/v{ONIGURUMA_VERSION}.tar.gz"

ONIGURUMA_TAR_GZ = DEPS_PATH / "oniguruma.tar.gz"
ONIGURUMA_BUILD = DEPS_PATH / "oniguruma_build"


DEFINITIONS = (pathlib.Path("onig") / "_onig_cffi" / "definitions.c").read_text()
SOURCE = (pathlib.Path("onig") / "_onig_cffi" / "source.c").read_text()


ffi = cffi.FFI()
ffi.cdef(DEFINITIONS)
ffi.set_source("onig._onig_cffi._onig_cffi", SOURCE)

extension = ffi.distutils_extension()


def download_oniguruma():
    DEPS_PATH.mkdir(exist_ok=True)
    if not ONIGURUMA_TAR_GZ.exists():
        response = request.urlopen(ONIGURUMA_URL)
        if response.status != 200:
            raise Exception(
                "unable to download the Oniguruma regular expression library"
            )
        ONIGURUMA_TAR_GZ.write_bytes(response.read())


def build_oniguruma(environment):
    if ONIGURUMA_BUILD.exists():
        shutil.rmtree(ONIGURUMA_BUILD)
    with tarfile.open(ONIGURUMA_TAR_GZ, "r:gz") as oniguruma_tar_gz:
        oniguruma_tar_gz.extractall(path=ONIGURUMA_BUILD)
    oniguruma_path = next(ONIGURUMA_BUILD.glob("oniguruma-*"))
    extension.include_dirs.append(str(oniguruma_path / "src"))
    if sys.platform == "win32":
        subprocess.check_call(
            ["make_win.bat"], shell=True, cwd=oniguruma_path, env=environment
        )
        extension.libraries.append("onig_s")
        extension.library_dirs.append(str(oniguruma_path))

        # shutil.copy(oniguruma_path / "onig.dll", "onig/_onig_cffi/onig.dll")
    elif sys.platform == "linux":
        subprocess.check_call(
            ["autoreconf", "-vfi"], cwd=oniguruma_path, env=environment
        )
        subprocess.check_call(["./configure"], cwd=oniguruma_path, env=environment)
        subprocess.check_call(["make"], cwd=oniguruma_path, env=environment)

        extension.extra_objects.append(
            str(oniguruma_path / "src" / ".libs" / "libonig.a")
        )
    else:
        raise Exception(f"cannot build Oniguruma for platform {sys.platform!r}")


class BuildExtensions(build_ext):
    def build_extensions(self):
        DEPS_PATH.mkdir(exist_ok=True)

        download_oniguruma()

        if sys.platform == "win32":
            from distutils import _msvccompiler
            from distutils.util import get_platform

            platform_name = self.compiler.plat_name or get_platform()
            platform_spec = _msvccompiler.PLAT_TO_VCVARS[platform_name]

            vc_env = _msvccompiler._get_vc_env(platform_spec)

            # if "CL" not in vc_env:
            #     vc_env["CL"] = ""
            # vc_env["CL"] += " /DONIGURUMA_EXPORT="

            build_oniguruma(environment=vc_env)
        elif sys.platform == "linux":
            environ = dict(os.environ)
            if "CFLAGS" not in environ:
                environ["CFLAGS"] = ""
            environ["CFLAGS"] += " -fPIC"
            build_oniguruma(environment=environ)
        else:
            raise Exception(f"cannot build Oniguruma for platform {sys.platform!r}")

        build_ext.build_extensions(self)


def build(setup_kwargs):
    setup_kwargs.update(
        {"ext_modules": [extension], "cmdclass": {"build_ext": BuildExtensions}}
    )



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <mkoehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import abc
import enum
import platform
import warnings

from ._onig_cffi import ffi as _ffi, lib as _lib

try:
    from unibuf import UnicodeBuffer
except ImportError:
    if platform.python_implementation() == "CPython":
        warnings.warn(
            f"'unibuf' is not installed; 'onig' will be slow for unicode strings",
            category=ResourceWarning,
        )
    UnicodeBuffer = None


class EncodedString(abc.ABC):
    encoding: Encoding

    unicode_string: str
    encoded_string: bytes

    def __init__(
        self, encoding: Encoding, unicode_string: str, byte_string: bytes
    ) -> None:
        self.encoding = encoding
        self.unicode_string = unicode_string
        self.byte_string = byte_string

    @abc.abstractmethod
    def byte_offset_to_char_offset(self, offset: int) -> int:
        raise NotImplementedError()

    @abc.abstractmethod
    def char_offset_to_byte_offset(self, offset: int) -> int:
        raise NotImplementedError()

    def substring_by_byte_offset(self, start: int, end: int) -> str:
        return self.substring_by_char_offset(
            self.byte_offset_to_char_offset(start), self.byte_offset_to_char_offset(end)
        )

    def substring_by_char_offset(self, start: int, end: int) -> str:
        return self.unicode_string[start:end]


class SingleByteString(EncodedString):
    def byte_offset_to_char_offset(self, offset: int) -> int:
        return offset

    def char_offset_to_byte_offset(self, offset: int) -> int:
        return offset


class UTF8String(EncodedString):
    def byte_offset_to_char_offset(self, offset: int) -> int:
        return len(self.byte_string[:offset].decode("utf-8"))

    def char_offset_to_byte_offset(self, offset: int) -> int:
        return len(self.unicode_string[:offset].encode("utf-8"))

    def substring_by_byte_offset(self, start: int, end: int) -> str:
        return self.byte_string[start:end].decode("utf-8")


class UTF16String(EncodedString):
    def byte_offset_to_char_offset(self, offset: int) -> int:
        return offset // 2

    def char_offset_to_byte_offset(self, offset: int) -> int:
        return offset * 2


class UTF32String(EncodedString):
    def byte_offset_to_char_offset(self, offset: int) -> int:
        return offset // 4

    def char_offset_to_byte_offset(self, offset: int) -> int:
        return offset * 4


_encoding_map: t.Dict[str, Encoding] = {}


class Encoding(enum.Enum):
    encoding_name: str

    onig_encoding: t.Any

    index: int

    string_cls: t.Type[EncodedString]

    ASCII = "ASCII", _lib.OnigEncodingASCII, 0, SingleByteString

    ISO_8859_1 = "ISO-8859-1", _lib.OnigEncodingISO_8859_1, 1, SingleByteString

    UTF_8 = "UTF-8", _lib.OnigEncodingUTF8, 2, UTF8String

    UTF_16LE = "UTF-16LE", _lib.OnigEncodingUTF16_LE, 3, UTF16String
    UTF_16BE = "UTF-16BE", _lib.OnigEncodingUTF16_BE, 3, UTF16String

    UTF_32LE = "UTF-32LE", _lib.OnigEncodingUTF32_LE, 4, UTF32String
    UTF_32BE = "UTF-32BE", _lib.OnigEncodingUTF32_BE, 4, UTF32String

    def __init__(
        self,
        encoding_name: str,
        onig_encoding: t.Any,
        index: int,
        string_cls: t.Type[EncodedString],
    ) -> None:
        _encoding_map[encoding_name] = self
        self.encoding_name = encoding_name
        self.onig_encoding = onig_encoding
        self.index = index
        self.string_cls = string_cls

    def encode(self, text: str) -> EncodedString:
        return self.string_cls(self, text, text.encode(self.encoding_name))

    @staticmethod
    def get_by_name(name: str) -> Encoding:
        try:
            return _encoding_map[name.upper()]
        except KeyError:
            raise Exception(f"unknown encoding {name}")


class WrappedString:
    base_string: str
    native_encoding: Encoding

    _encoded_strings: t.Dict[Encoding, EncodedString]

    def __init__(
        self,
        base_string: str,
        native_encoding: Encoding = Encoding.UTF_8,
        *,
        _encoded_strings: t.Optional[t.Dict[Encoding, EncodedString]] = None,
    ) -> None:
        self.base_string = base_string
        self.native_encoding = native_encoding
        self._encoded_strings = _encoded_strings or {}

    def encode(self, encoding: Encoding) -> EncodedString:
        try:
            return self._encoded_strings[encoding]
        except KeyError:
            self._encoded_strings[encoding] = encoding.encode(self.base_string)
            return self._encoded_strings[encoding]


def wrap(
    base_string: str, *, native_encoding: t.Optional[Encoding] = None
) -> WrappedString:
    if native_encoding is None and UnicodeBuffer is not None:
        buffer = UnicodeBuffer(base_string)
        native_encoding = Encoding.get_by_name(buffer.encoding)
        return WrappedString(
            base_string,
            native_encoding,
            _encoded_strings={
                native_encoding: native_encoding.string_cls(
                    native_encoding, base_string, bytes(buffer)
                )
            },
        )
    return WrappedString(base_string, native_encoding or Encoding.UTF_8)


def _onig_initialize() -> None:
    encodings = _ffi.new(
        "OnigEncoding[]",
        [_ffi.addressof(encoding.onig_encoding) for encoding in Encoding],
    )
    _lib.onig_initialize(encodings, _ffi.sizeof(encodings) // _ffi.sizeof(encodings[0]))


_onig_initialize()



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <mkoehl@cs.uni-saarland.de>

from __future__ import annotations

from .pattern import Option, Pattern
from .match import Match, Region


__all__ = ["Option", "Pattern", "Match", "Region"]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <mkoehl@cs.uni-saarland.de>

from __future__ import annotations

import pathlib

from importlib import resources

from cffi import FFI


DEFINITIONS = resources.read_text(__package__, "definitions.c")
SOURCE = resources.read_text(__package__, "source.c")

builder = FFI()
builder.cdef(DEFINITIONS)

builder.set_source("onig._onig_cffi.onig", SOURCE, libraries=["onig"])


if __name__ == "__main__":
    builder.compile()



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <mkoehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import dataclasses
import functools

from . import _encoding


if t.TYPE_CHECKING:
    from .pattern import Pattern


@dataclasses.dataclass(frozen=True)
class Region:
    begin: int
    end: int


Regions = t.Sequence[Region]


class Match:
    pattern: Pattern

    _encoded_string: _encoding.EncodedString
    _byte_regions: Regions

    def __init__(
        self,
        pattern: Pattern,
        encoded_string: _encoding.EncodedString,
        byte_regions: Regions,
    ) -> None:
        self.pattern = pattern
        self._encoded_string = encoded_string
        self._byte_regions = byte_regions

    @functools.cached_property
    def start(self) -> int:
        return self._encoded_string.byte_offset_to_char_offset(
            self._byte_regions[0].begin
        )

    @functools.cached_property
    def end(self) -> int:
        return self._encoded_string.byte_offset_to_char_offset(
            self._byte_regions[0].end
        )

    @functools.cached_property
    def region(self) -> Region:
        return Region(self.start, self.end)

    @functools.cached_property
    def regions(self) -> Regions:
        return tuple(
            Region(
                self._encoded_string.byte_offset_to_char_offset(region.begin),
                self._encoded_string.byte_offset_to_char_offset(region.end),
            )
            for region in self._byte_regions
        )

    @t.overload
    def group(self, key: int) -> str:
        pass

    @t.overload
    def group(self, key: str) -> t.Sequence[str]:
        pass

    def group(self, key: t.Union[int, str]) -> t.Union[str, t.Sequence[str]]:
        if isinstance(key, int):
            return self._group_by_number(key)
        else:
            return [
                self._group_by_number(index)
                for index in self.pattern.get_group_numbers(key)
            ]

    def _group_by_number(self, index: int) -> str:
        try:
            return self._encoded_string.substring_by_byte_offset(
                self._byte_regions[index].begin, self._byte_regions[index].end
            )
        except IndexError:
            raise Exception(f"invalid group number {index}")

    def groups(self) -> t.Sequence[str]:
        return [
            self._group_by_number(index) for index in range(len(self._byte_regions))
        ]

    def __repr__(self) -> str:
        return (
            f"<onig.Match region=({self.start}, {self.end}), match={self.group(0)!r}>"
        )



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <mkoehl@cs.uni-saarland.de>

from __future__ import annotations

try:
    from ._onig_cffi import ffi, lib
except ImportError:
    from .build import builder

    builder.compile()

    from ._onig_cffi import ffi, lib


__all__ = ["ffi", "lib"]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <mkoehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import enum

from ._onig_cffi import ffi as _ffi, lib as _lib

from . import _encoding, match


class Option(enum.IntFlag):
    NONE = _lib.ONIG_OPTION_NONE

    SINGLELINE = _lib.ONIG_OPTION_SINGLELINE
    MULTILINE = _lib.ONIG_OPTION_MULTILINE

    IGNORECASE = _lib.ONIG_OPTION_IGNORECASE

    EXTEND = _lib.ONIG_OPTION_EXTEND

    FIND_LONGEST = _lib.ONIG_OPTION_FIND_LONGEST
    FIND_NOT_EMPTY = _lib.ONIG_OPTION_FIND_NOT_EMPTY

    NEGATE_SINGLELINE = _lib.ONIG_OPTION_NEGATE_SINGLELINE


class _DummyType:
    def __new__(cls, *args, **kwargs):
        raise Exception("only for typing")


_OnigRegex = t.NewType("_OnigRegex", _DummyType)


def _compile_pattern(
    pattern: _encoding.WrappedString, encoding: _encoding.Encoding, options: Option
) -> _OnigRegex:
    encoded_string = pattern.encode(encoding)
    regex = _ffi.new("OnigRegex*")
    error = _ffi.new("OnigErrorInfo*")
    code = _lib.onig_cffi_new(
        regex,
        encoded_string.byte_string,
        len(encoded_string.byte_string),
        options,
        _ffi.addressof(encoding.onig_encoding),
        error,
    )
    if code != _lib.ONIG_NORMAL:
        error_message = _ffi.new(f"OnigUChar[{_lib.ONIG_MAX_ERROR_MESSAGE_LEN}]")
        _lib.onig_error_code_to_str(error_message, code, error)
        raise Exception(_ffi.string(error_message).decode("ascii"))
    return regex


def _new_region() -> _OnigRegion:
    return _ffi.gc(_lib.onig_region_new(), _lib.onig_cffi_region_free)


class Pattern:
    _pattern: _encoding.WrappedString
    _regexes: t.Dict[_encoding.Encoding, _OnigRegex]
    _options: Option

    def __init__(self, pattern: str, *, options: Option = Option.NONE) -> None:
        self._pattern = _encoding.wrap(pattern)
        self._regexes = {}
        self._options = options
        self._compile(self._pattern.native_encoding)

    @property
    def pattern(self) -> str:
        return self._pattern.base_string

    @property
    def options(self) -> Option:
        return self._options

    def _compile(self, encoding: _encoding.Encoding) -> _OnigRegex:
        try:
            return self._regexes[encoding]
        except KeyError:
            self._regexes[encoding] = _compile_pattern(
                self._pattern, encoding, self._options
            )
            return self._regexes[encoding]

    def _get_common_encoding(
        self, wrapped_text: _encoding.WrappedString
    ) -> _encoding.Encoding:
        if wrapped_text.native_encoding.index > self._pattern.native_encoding.index:
            return wrapped_text.native_encoding
        else:
            return self._pattern.native_encoding

    def _search(
        self, encoded_text: _encoding.EncodedString, offset: int = 0
    ) -> t.Optional[match.Match]:
        region = _new_region()
        position = _lib.onig_cffi_search(
            self._compile(encoded_text.encoding),
            encoded_text.byte_string,
            len(encoded_text.byte_string),
            offset,
            region,
            _lib.ONIG_OPTION_NONE,
        )
        regions = [
            match.Region(region.beg[index], region.end[index])
            for index in range(region.num_regs)
        ]
        if position >= 0:
            return match.Match(self, encoded_text, regions)
        return None

    def _prepare(
        self, text: str, *, start: int = 0
    ) -> t.Tuple[_encoding.EncodedString, int]:
        wrapped_text = _encoding.wrap(text)
        encoding = self._get_common_encoding(wrapped_text)
        encoded_text = wrapped_text.encode(encoding)
        offset = encoded_text.char_offset_to_byte_offset(start)
        return encoded_text, offset

    def finditer(self, text: str, *, start: int = 0) -> t.Iterator[match.Match]:
        encoded_text, offset = self._prepare(text, start=start)
        while offset < len(encoded_text.byte_string):
            match = self._search(encoded_text, offset)
            if match is None:
                break
            yield match
            offset = match._byte_regions[0].end

    def findall(self, text: str, *, start: int = 0) -> t.Sequence[match.Match]:
        return list(self.finditer(text, start=start))

    def search(self, text: str, *, start: int = 0) -> t.Optional[match.Match]:
        encoded_text, offset = self._prepare(text, start=start)
        return self._search(encoded_text, offset)

    def get_group_numbers(self, name: str) -> t.Sequence[int]:
        encoding, regex = next(iter(self._regexes.items()))
        encoded_name = encoding.encode(name)
        number_list = _ffi.new("int**")
        list_length = _lib.onig_cffi_name_to_group_numbers(
            regex, encoded_name.byte_string, len(encoded_name.byte_string), number_list,
        )
        return [number_list[0][index] for index in range(list_length)]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['onig', 'onig._onig_cffi']

package_data = \
{'': ['*']}

install_requires = \
['cffi>=1.14.0,<2.0.0']

extras_require = \
{':implementation_name == "cpython"': ['unibuf>=0.1.1,<0.2.0']}

setup_kwargs = {
    'name': 'onig',
    'version': '0.0.7',
    'description': 'A Python wrapper around the Oniguruma regular expression library.',
    'long_description': 'Python Onig\n===========\n\nPython bindings for the Oniguruma regular expression library.\n',
    'author': 'Maximilian Köhl',
    'author_email': 'mkoehl@cs.uni-saarland.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://dgit.cs.uni-saarland.de/koehlma/onig',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
